@@ -4,38 +4,20 @@ my $debuglevel=0;
 use Carp;
 
 use strict;
+use Email::Abstract;
 use File::Basename;
 use Mail::Internet;
 use Sys::Hostname; (my $HOSTNAME = hostname) =~ s/\..*//;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $ASSUME_MSGPREFIX);
 use Fcntl ':flock';
-@ISA = qw(Mail::Internet);
-
 
 $ASSUME_MSGPREFIX = 0;
 
-# stolen from linux sysexits.h, YMMV on other OSes.  sorry, but it was either this or forcing everyone to h2ph.
-use constant EX_USAGE       => 64; # command line usage error
-use constant EX_DATAERR     => 65; # data format error
-use constant EX_NOINPUT     => 66; # cannot open input
-use constant EX_NOUSER      => 67; # addressee unknown
-use constant EX_NOHOST      => 68; # host name unknown
-use constant EX_UNAVAILABLE => 69; # service unavailable
-use constant EX_SOFTWARE    => 70; # internal software error
-use constant EX_OSERR       => 71; # system error (e.g., can't fork)
-use constant EX_OSFILE      => 72; # critical OS file missing
-use constant EX_CANTCREAT   => 73; # can't create (user) output file
-use constant EX_IOERR       => 74; # input/output error
-use constant EX_TEMPFAIL    => 75; # temp failure; user is invited to retry
-use constant EX_PROTOCOL    => 76; # remote error in protocol
-use constant EX_NOPERM      => 77; # permission denied
-use constant EX_CONFIG      => 78; # configuration error
-
-use constant DEFERRED  => EX_TEMPFAIL;
+use constant DEFERRED  => 75;
 use constant REJECTED  => 100;
 use constant DELIVERED => 0;
 
-$VERSION = '0.22';
+$VERSION = '0.301';
 
 =head1 NAME
 
@@ -57,7 +39,8 @@ nearly all of the same features, and more, and better.  Use it instead.
 =cut
 
 sub _debug {
-    my ($priority, $what) = @_; 
+    my ($self, $priority, $what) = @_; 
+    return $self->{logger}->($priority, $what) if $self->{logger};
     return if $debuglevel < $priority;
     chomp $what; chomp $what;
     my ($subroutine) = (caller(1))[3]; $subroutine =~ s/(.*):://;
@@ -99,38 +82,39 @@ goes in the worst case scenario.
 =cut
 
 sub new { 
-    my $class = shift;
-    my $stuff = shift;
+  my $class = shift;
+  my $stuff = shift;
 
-    my %opts = @_;
+  my %opts = @_;
 
-    my $self;
+  my $self = {
+    interpolate_strftime => 0,
+    one_for_all          => 0,
+    %opts,
+  };
 
-    # What sort of stuff do we have?
-    if (ref $stuff eq "Mail::Internet" or ref $stuff eq "MIME::Entity"){ 
-        $self = $stuff;
-    } elsif (ref $stuff eq "ARRAY" or ref $stuff eq "GLOB") { 
-        $self = Mail::Internet->new($stuff);
-    } else { 
-        croak "Data was neither a mail object or a reference to something I understand";
-    }
-    $self->{opts} = \%opts;
-    $self->{opts}->{interpolate_strftime} ||= 0;
-    $self->{opts}->{one_for_all}          ||= 0;
+  # What sort of stuff do we have?
+  if (eval { $stuff->isa('Mail::Internet') }) {
+    $self->{email} = $stuff;
+  } elsif (my $email = eval { Email::Abstract->new($self) }) {
+    $self->{email} = $email->cast('Mail::Internet');
+  } elsif (ref $stuff eq "ARRAY" or ref $stuff eq "GLOB") { 
+    $self->{email} = Mail::Internet->new($stuff);
+  } else { 
+    croak "Data was neither a mail object or a reference to something I understand";
+  }
 
-    my $default_unixbox = (grep { -d $_ } qw(/var/spool/mail/ /var/mail/))[0] 
-                        . getpwuid($>);
-    my $default_maildir = ((getpwuid($>))[7]) . "/Maildir/";
+  my $default_maildir = ((getpwuid($>))[7]) . "/Maildir/";
 
-    my $default_mbox = 
-        $ENV{MAIL} 
-        || (-e $default_unixbox && $default_unixbox)
-        || (-d $default_maildir . "cur" && $default_maildir);
+  my $default_mbox = 
+    $ENV{MAIL} 
+    || (-d File::Spec->catdir($default_maildir, 'new') ? $default_maildir : ())
+    || ((grep { -d $_ } qw(/var/spool/mail/ /var/mail/))[0] . getpwuid($>));
 
-    $self->{default_mbox} = $default_mbox;
-    $self->{opts}->{emergency} ||= $default_mbox;
+  $self->{default_mbox} = $default_mbox;
+  $self->{emergency}  ||= $default_mbox;
 
-    return bless $self, $class;
+  return bless $self => $class;
 }
 
 =head2 C<deliver($where, ...)>
@@ -193,7 +177,7 @@ sub _nifty_interpolate { # perform ~user and %Y%m%d strftime interpolation
     my $self = shift;
     my @out = @_;
     my @localtime = localtime;
-    if ($self->{opts}->{interpolate_strftime} and grep { /%/ } @out) {
+    if ($self->{interpolate_strftime} and grep { /%/ } @out) {
         require POSIX;
         import POSIX qw(strftime);
         @out = map { strftime($_, @localtime) } @out;
@@ -216,7 +200,7 @@ sub deliver {
 
     my @actually_saved_to_files = ();
 
-    _debug(2,"delivering to @files");
+    $self->_debug(2, "delivering to @files");
 
     # from man procmailrc:
     #         If  it  is  a  directory,  the mail will be delivered to a newly
@@ -246,18 +230,19 @@ sub deliver {
     for my $file (@files) {
         my $mailbox_type = $self->_mailbox_type($file);
         push @{$deliver_types{$mailbox_type}}, $file;
-        _debug(3, "$file is of type $mailbox_type");
+        $self->_debug(3, "$file is of type $mailbox_type");
     }
 
     foreach my $deliver_type (sort keys %deliver_types) {
         next if not @{$deliver_types{$deliver_type}};
-        my $deliver_handler = "deliver_to_$deliver_type";
-        _debug(3, "calling deliver handler $deliver_handler(@{$deliver_types{$deliver_type}})");
+        my $deliver_handler = "_deliver_to_$deliver_type";
+        $self->_debug(3, "calling deliver handler $deliver_handler(@{$deliver_types{$deliver_type}})");
 
         # Don't try to deliver to things for which we have no delivery method.
         next unless $self->can($deliver_handler);
 
-        push @actually_saved_to_files, $self->$deliver_handler(@{$deliver_types{$deliver_type}});
+        push @actually_saved_to_files,
+          $self->$deliver_handler(@{$deliver_types{$deliver_type}});
     }
 
     if (@actually_saved_to_files == 0) {
@@ -269,15 +254,15 @@ sub deliver {
         # because we want to be able to trap the die message inside an eval {}
         # for testing purposes.
 
-        my $emergency = $self->{opts}->{emergency};
+        my $emergency = $self->{emergency};
         if (not defined $emergency) {
-            return ();
+            return;
         } else {
             if (grep ($emergency eq $_, @files)) { # already tried that mailbox
                 return ();
             } else {
                 my $deliver_type = $self->_mailbox_type($emergency);
-                my $deliver_handler = "deliver_to_$deliver_type";
+                my $deliver_handler = "_deliver_to_$deliver_type";
                 @actually_saved_to_files = $self->$deliver_handler($emergency);
                 return if not @actually_saved_to_files;
             }
@@ -297,8 +282,8 @@ sub _mailbox_type {
 
     if (-d $file) {
         return 'maildir' if (-d "$file/tmp" and -d "$file/new");
-        if (exists($self->{opts}->{ASSUME_MSGPREFIX})) {
-            return 'msgprefix' if $self->{opts}->{ASSUME_MSGPREFIX};
+        if (exists($self->{ASSUME_MSGPREFIX})) {
+            return 'msgprefix' if $self->{ASSUME_MSGPREFIX};
             return "maildir";
         }
         return 'msgprefix' if $ASSUME_MSGPREFIX;
@@ -307,15 +292,33 @@ sub _mailbox_type {
     return 'mbox';
 }
 
+sub _get_opt {
+  my ($self, $arg) = @_;
+
+  my $opt;
+
+  if (ref $arg->[0] eq 'HASH') {
+    Carp::carp "prepending arguments is deprecated; append them instead"
+      unless @$arg == 1;
+    $opt = shift @$arg;
+  } elsif (ref $arg->[-1] eq 'HASH') {
+    $opt = pop @$arg;
+  }
+
+  return $opt || {};
+}
+
 # ----------------------------------------------------------
 sub _deliver_to_mbox {
 # ----------------------------------------------------------
     my $self = shift;
+    my $opt  = $self->_get_opt(\@_);
+
     my @saved_to = ();
     foreach my $file (@_) {
         # auto-create the parent dir.
-        if (my $mkdir_error = _mkdir_p(dirname($file))) {
-          _debug(0, $mkdir_error);
+        if (my $mkdir_error = $self->_mkdir_p(dirname($file))) {
+          $self->_debug(0, $mkdir_error);
           next;
         }
         my $error = $self->_write_message(
@@ -327,7 +330,7 @@ sub _deliver_to_mbox {
           }
         );
         if (not $error) { push @saved_to, $file; }
-        else            { _debug(1, $error); }
+        else            { $self->_debug(1, $error); }
     }
     return @saved_to;
 }
@@ -343,39 +346,42 @@ sub _write_message {
     $write_opts->{need_lock} = 1 if not defined $write_opts->{need_lock};
     $write_opts->{extra_newline} = 0 if not defined $write_opts->{extra_newline};
 
-    _debug(3, "writing to $file; options @{[%$write_opts]}");
+    $self->_debug(3, "writing to $file; options @{[%$write_opts]}");
 
     unless (open(FH, ">>$file")) { return "Couldn't open $file: $!"; }
 
     if ($write_opts->{need_lock}) {
-      my $lock_error = _audit_get_lock(\*FH, $file);
+      my $lock_error = $self->_audit_get_lock(\*FH, $file);
       return $lock_error if $lock_error;
     }
 
     seek FH, 0, 2;
 
-    if (not $write_opts->{need_from} and $self->head->header->[0] =~ /^From\s/)
+    if (not $write_opts->{need_from} and $self->{email}->head->header->[0] =~ /^From\s/)
     {
-        _debug(3,"mbox From line found, stripping because we're maildir");
-        $self->head->delete("From ");
-        $self->unescape_from();
+        $self->_debug(3,"mbox From line found, stripping because we're maildir");
+        $self->{email}->head->delete("From ");
+        $self->{email}->unescape_from;
     }
 
-    if ($write_opts->{'need_from'} and $self->head->header->[0] !~ /^From\s/) {
-        _debug(3,"No mbox From line, making one up.");
+    if ($write_opts->{need_from} and $self->{email}->head->header->[0] !~ /^From\s/) {
+        $self->_debug(3,"No mbox From line, making one up.");
         if (exists $ENV{UFLINE}) {
-            _debug(3,"Looks qmail, but preline not run, prepending UFLINE, RPLINE, DTLINE");
+            $self->_debug(3,"Looks qmail, but preline not run, prepending UFLINE, RPLINE, DTLINE");
             print FH $ENV{UFLINE};
             print FH $ENV{RPLINE};
             print FH $ENV{DTLINE};
         } else {
             my $from = (
-              $self->get('Return-path') ||
-              $self->get('Sender')      ||
-              $self->get('Reply-To')    ||
+              $self->{email}->get('Return-path') ||
+              $self->{email}->get('Sender')      ||
+              $self->{email}->get('Reply-To')    ||
+              # It might not be insane to include From header here.
+              # -- rjbs, 2006-07-25
               'root@localhost'
             );
             chomp $from;
+
             # comment <email@address> -> email@address
             $from = $1 if $from =~ /<(.*?)>/;
 
@@ -392,17 +398,19 @@ sub _write_message {
         }
     }
 
-    _debug(4, "printing self as mbox string.");
-    print FH $self->as_string;
+    $self->_debug(4, "printing self as mbox string.");
+    print FH $self->{email}->as_string or return "couldn't write to $file: $!";
+
     # extra \n added because mutt seems to like a "\n\nFrom " in mbox files
-    print FH "\n" if $write_opts->{extra_newline};
+    print FH "\n" if $write_opts->{extra_newline}
+      or return "couldn't write to $file: $!";
 
     if ($write_opts->{need_lock}) {
         flock(FH, LOCK_UN) or return "Couldn't unlock $file";
     }
 
     close FH or return "Couldn't close $file after writing: $!";
-    _debug(4, "returning success.");
+    $self->_debug(4, "returning success.");
     return 0; # success
 }
 
@@ -410,8 +418,8 @@ sub _write_message {
 # NOT IMPLEMENTED
 # ----------------------------------------------------------
 
-sub _deliver_to_mh        { my $self = shift; my @saved_to=(); } 
-sub _deliver_to_msgprefix { my $self = shift; my @saved_to=(); }
+#sub _deliver_to_mh        { my $self = shift; my @saved_to=(); } 
+#sub _deliver_to_msgprefix { my $self = shift; my @saved_to=(); }
 
 # variables for _deliver_to_maildir
 
@@ -424,30 +432,33 @@ sub _deliver_to_maildir {
     my $self = shift;
     my @saved_to = ();
 
-    _debug(3, "will write to @_");
+    $self->_debug(3, "will write to @_");
 
     # since mutt won't add a lines tag to maildir messages, we'll add it here
-    unless (length $self->get("Lines")) {
-        my $num_lines = @{$self->body};
-        $self->head->add("Lines", $num_lines);
-        _debug(4,"Adding Lines: $num_lines header");
+    unless (length ($self->{email}->get("Lines")||'')) {
+        my $num_lines = @{$self->{email}->body};
+        $self->{email}->head->add("Lines", $num_lines);
+        $self->_debug(4,"Adding Lines: $num_lines header");
     }
 
-    if ($maildir_time != time) { $maildir_time = time; $maildir_counter = 0 } else { $maildir_counter++ }
+    if ($maildir_time != time) {
+        $maildir_time = time;
+        $maildir_counter = 0;
+    } else {
+        $maildir_counter++;
+    }
 
     # write the tmp file.
     # hardlink to all the new files.
     # unlink the temp file.
 
-    # 
     # write the tmp file in the first writable maildir directory.
-    # 
 
     my $tmp_path;
     foreach my $file (my @maildirs = @_) {
 
         $file =~ s/\/$//;
-        my $tmpdir = $self->{opts}->{"one_for_all"} ? $file : "$file/tmp";
+        my $tmpdir = $self->{one_for_all} ? $file : "$file/tmp";
 
         my $msg_file;
         do {
@@ -455,21 +466,31 @@ sub _deliver_to_maildir {
         } while ( -e "$tmpdir/$msg_file" );
 
         $tmp_path = "$tmpdir/$msg_file";
-        _debug(3,"writing to $tmp_path");
+        $self->_debug(3,"writing to $tmp_path");
 
         # auto-create the maildir.
-        if (my $mkdir_error = _mkdir_p(
-                                      $self->{opts}->{"one_for_all"}
-                                      ? ($file)
-                                      : map { "$file/$_" } qw(tmp new cur))) { _debug(0, $mkdir_error); next; }
+        if (my $mkdir_error
+          = $self->_mkdir_p($self->{one_for_all} ? ($file)
+                                          : map { "$file/$_" } qw(tmp new cur))
+        ) {
+          $self->_debug(0, $mkdir_error);
+          next;
+        }
 
-        my $error = $self->_write_message($tmp_path, {need_from=>0, need_lock=>0});
-        if (not $error) { last; }  # only write to the first writeable maildir
-        else            { _debug(1, $error);
-                          unlink $tmp_path;
-                          $tmp_path = undef;
-                          next;
-                      }
+        my $error = $self->_write_message(
+          $tmp_path,
+          {
+            need_from => 0,
+            need_lock => 0
+          }
+        );
+
+        last unless $error; # only write to the first writeable maildir
+
+        $self->_debug(1, $error);
+        unlink $tmp_path;
+        $tmp_path = undef;
+        next;
     }
 
     if (not $tmp_path) { return 0 } # unable to write to any of the specified maildirs.
@@ -482,36 +503,36 @@ sub _deliver_to_maildir {
         $file =~ s/\/$//;
 
         my $msg_file;
-        my $newdir = $self->{opts}->{"one_for_all"} ? $file : "$file/new";
+        my $newdir = $self->{one_for_all} ? $file : "$file/new";
         $maildir_counter = 0;
         do {
             $msg_file = join ".", ($maildir_time=time, $$ . "_$maildir_counter", $HOSTNAME); $maildir_counter++;
         } while ( -e "$newdir/$msg_file" );
 
         # auto-create the maildir.
-        if (my $mkdir_error = _mkdir_p(
-                                      $self->{opts}->{"one_for_all"}
+        if (my $mkdir_error = $self->_mkdir_p(
+                                      $self->{one_for_all}
                                       ? ($file)
-                                      : map { "$file/$_" } qw(tmp new cur))) { _debug(0, $mkdir_error); next; }
+                                      : map { "$file/$_" } qw(tmp new cur))) { $self->_debug(0, $mkdir_error); next; }
 
         my $new_path = "$newdir/$msg_file";
-        _debug(3,"maildir: hardlinking to $new_path");
+        $self->_debug(3,"maildir: hardlinking to $new_path");
 
         if    (link $tmp_path, $new_path) { push @saved_to, $new_path; }
         else {
             require Errno; import Errno qw(EXDEV);
             if ($! == &EXDEV) { # Invalid cross-device link, see /usr/**/include/*/errno.h
-                _debug(0,"Couldn't link $tmp_path to $new_path: $!");
-                _debug(0,"attempting direct maildir delivery to $new_path...");
+                $self->_debug(0,"Couldn't link $tmp_path to $new_path: $!");
+                $self->_debug(0,"attempting direct maildir delivery to $new_path...");
                 push @saved_to, $self->_deliver_to_maildir($file);
                 next;
             }
-            else { _debug(0,"Couldn't link $tmp_path to $new_path: $!"); }
+            else { $self->_debug(0,"Couldn't link $tmp_path to $new_path: $!"); }
         }
     }
 
     # unlink the temp file
-    unlink $tmp_path or _debug(1,"Couldn't unlink $tmp_path: $!");
+    unlink $tmp_path or $self->_debug(1,"Couldn't unlink $tmp_path: $!");
     return @saved_to;
 }
 # ----------------------------------------------------------
@@ -519,32 +540,30 @@ sub _deliver_to_maildir {
 # ----------------------------------------------------------
 
 sub _audit_get_lock {
+    my $self = shift;
     my $FH   = shift;
     my $file = shift;
-    _debug(4, "  attempting to lock  file $file");
+    $self->_debug(4, "  attempting to lock  file $file");
     for (1..10) {
-        if (flock($FH, LOCK_EX)) { _debug(4, "  successfully locked file $file"); return; }
+        if (flock($FH, LOCK_EX)) { $self->_debug(4, "  successfully locked file $file"); return; }
         else                     { sleep $_ and next; }
     }
-    _debug(1,my $errstr="Couldn't get exclusive lock on $file");
+    $self->_debug(1,my $errstr="Couldn't get exclusive lock on $file");
     return $errstr;
 }
 
 sub _mkdir_p { # mkdir -p (also create parents if necessary)
-    return if not @_;
-    return if not length $_[0];
-    foreach (@_) {
-        next if -d $_;
-        while (/\/$/) { chop }
-        _debug(4, "$_ doesn't exist, creating.");
-        if (my $error = _mkdir_p(dirname($_))) { return $error }
-        mkdir ($_, 0777) or return "unable to mkdir $_: $!";
+    my $self = shift;
+    return unless @_;
+    foreach my $path (@_) {
+        next if -d $path;
+        $self->_debug(4, "$path doesn't exist, creating.");
+        eval { File::Path::mkpath($path, 0, 0755) };
+        return $@ if $@;
     }
     return;
 }
 
-sub _myALRM { die "alarm\n" }
-
 1;
 __END__
 
